!---------------------------------------------------------------------------------------
! PREP_SATELLITE: mod_settings
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2022, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_settings
!!
!! Purpose:    Module for the construction of the settings data structure.
!!
!---------------------------------------------------------------------------------------

module mod_settings

  use mod_var

  implicit none
  private

  public :: settings_t, read_settings, split, parse_string

  ! settings_t contains all settings

  type :: settings_t

    character(len=max_name_len)     :: satellite
    character(len=max_name_len)     :: satname
    character(len=max_path_len)     :: path_flexpart
    character(len=max_path_len)     :: path_options
    character(len=max_path_len)     :: path_output
    character(len=max_path_len)     :: path_ohfield
    character(len=max_path_len)     :: path_obs
    character(len=max_path_len)     :: path_obsout
    character(len=max_name_len)     :: suffix
    character(len=max_path_len)     :: file_avail
    character(len=max_path_len)     :: file_availnest
    character(len=max_name_len)     :: windfield

    logical                         :: landonly
    integer                         :: opmode_sel
    character(len=max_name_len)     :: nmeas_name
    character(len=max_name_len)     :: npixel_name
    character(len=max_name_len)     :: ncoord_name
    character(len=max_name_len)     :: nlevel_name
    character(len=max_name_len)     :: time_name
    character(len=max_name_len)     :: tdel_name
    character(len=max_name_len)     :: qa_name
    character(len=max_name_len)     :: oper_name
    real                            :: qa_cutoff
    character(len=max_name_len)     :: vmr_name
    character(len=max_name_len)     :: verr_name
    character(len=max_name_len)     :: lat_name
    character(len=max_name_len)     :: lon_name
    character(len=max_name_len)     :: ak_name
    character(len=max_name_len)     :: apri_name
    character(len=max_name_len)     :: dryair_name
    character(len=max_name_len)     :: psurf_name
    character(len=max_name_len)     :: pdel_name
    character(len=max_name_len)     :: surfclass_name
    character(len=max_name_len)     :: landfrac_name

    integer                         :: datei
    integer                         :: datef
    integer                         :: outrate
    integer                         :: outaverage
    integer                         :: outsample
    integer                         :: ind_source
    integer                         :: ind_receptor
    integer                         :: lnested
    integer                         :: linit_cond

    real                            :: outlonleft
    real                            :: outlatlower
    integer                         :: numxgrid
    integer                         :: numygrid
    integer                         :: numzgrid
    real                            :: dxout
    real                            :: dyout
    real                            :: outlonnest
    real                            :: outlatnest
    integer                         :: numxnest
    integer                         :: numynest
    real                            :: dxoutnest
    real                            :: dyoutnest
    real, dimension(:), allocatable :: zlevel

    character(len=10)               :: species
    integer                         :: ageclass
    integer                         :: npart
    integer                         :: mass
    logical                         :: avg_pixels
    logical                         :: usegridfile
    real                            :: dmin
    integer                         :: nsteps
    real                            :: cutoff
    integer                         :: nmin
    character(len=max_path_len)     :: filegrid

  end type settings_t

  contains

    ! --------------------------------------------------
    ! check if input is numeric
    ! --------------------------------------------------

    logical function is_numeric(string)

      character(len=200), intent(in) :: string
      real :: var
      integer :: ierr

      read(string,*,iostat=ierr) var
      if(ierr.eq.0) then
        is_numeric=.true.
      else
        is_numeric=.false.
      endif

    end function is_numeric

    ! --------------------------------------------------
    ! check if input is logical
    ! --------------------------------------------------

    logical function is_logical(string)

      character(len=200), intent(in) :: string
      logical :: var
      integer :: ierr

      read(string,*,iostat=ierr) var
      if(ierr.eq.0) then
        is_logical=.true.
      else
        is_logical=.false.
      endif

    end function is_logical

    ! --------------------------------------------------
    ! split string
    ! --------------------------------------------------

    subroutine split(str,delims,before,sep)

    ! Routine finds the first instance of a character from 'delims' in the
    ! the string 'str'. The characters before the found delimiter are
    ! output in 'before'. The characters after the found delimiter are
    ! output in 'str'. The optional output character 'sep' contains the 
    ! found delimiter. A delimiter in 'str' is treated like an ordinary 
    ! character if it is preceded by a backslash (\). If the backslash 
    ! character is desired in 'str', then precede it with another backslash.

    character(len=*) :: str, delims, before
    character,optional :: sep
    logical :: pres
    character :: ch,cha
    integer :: i, k, ibsl, ipos, iposa, lenstr

    pres=present(sep)
    str=adjustl(str)
    lenstr=len_trim(str)
    if(lenstr == 0) return        ! string str is empty
    k=0
    ibsl=0                        ! backslash initially inactive
    before=' '
    do i=1,lenstr
      ch=str(i:i)
      if(ibsl == 1) then          ! backslash active
        k=k+1
        before(k:k)=ch
        ibsl=0
        cycle
      end if
      if(ch == '\') then          ! backslash with backslash inactive
        k=k+1
        before(k:k)=ch
        ibsl=1
        cycle
      end if
      ipos=index(delims,ch)
      if(ipos == 0) then          ! character is not a delimiter
        k=k+1
        before(k:k)=ch
        cycle
      end if
      if(ch /= ' ') then          ! character is a delimiter that is not a space
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
      cha=str(i+1:i+1)            ! character is a space delimiter
      iposa=index(delims,cha)
      if(iposa > 0) then          ! next character is a delimiter
        str=str(i+2:)
        if(pres) sep=cha
        exit
      else
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
    end do
    if(i >= lenstr) str=''
    str=trim(str)
    before=trim(before)
    return

    end subroutine split

    ! --------------------------------------------------
    ! routine to identify and read in content
    ! --------------------------------------------------

    subroutine read_content (line, identifier, cc, cn, cl, match)

      character (len=200), intent (in) :: line, identifier
      character (len=200), intent (out) :: cc ! character content
      real(kind=8),        intent (out) :: cn ! numeric content
      logical,             intent (out) :: cl ! logical content
      logical,             intent (out) :: match

      integer :: n

      n = len_trim (identifier)

      ! default: line does not match, cc="", cn=-9900.0
      cc    = ""
      cn    = -9999.9
      match = .false.
      if ( len_trim (line) >= n ) then
        if ( line(:n) == identifier(:n) ) then
          cc = adjustl (line(n+1:))
          if ( is_numeric(cc) ) then
            read (line(n+1:),*) cn
          else
            if( is_logical(cc) ) read (line(n+1:),*) cl
          endif
          match = .true.
        end if
      end if

    end subroutine read_content

    ! --------------------------------------------------
    ! routine to parse a string
    ! --------------------------------------------------

    subroutine parse_string (str, delims, args, nargs)

    ! Parses the string 'str' into arguments args(1), ..., args(nargs) based on
    ! the delimiters contained in the string 'delims'. Preceding a delimiter in
    ! 'str' by a backslash (\) makes this particular instance not a delimiter.
    ! The integer output variable nargs contains the number of arguments found.

    character(len=*) :: str, delims
    character(len=len_trim(str)) :: strsav
    character(len=*), dimension(:) :: args
    integer :: i, na, lenstr, nargs

    strsav=str
    na=size(args)

    do i=1,na
      args(i)=' '
    end do
    nargs=0
    lenstr=len_trim(str)
    if(lenstr==0) return

    do
      if(len_trim(str) == 0) exit
      nargs=nargs+1
      call split(str,delims,args(nargs))
    end do
    str=strsav

    end subroutine parse_string

    ! --------------------------------------------------
    ! read settings
    ! --------------------------------------------------

    subroutine read_settings(filename, settings)

      character(len=200), intent(in)     :: filename
      type(settings_t),   intent(in out) :: settings

      integer            :: ierr, i, n
      character(len=200) :: line, identifier, cc
      real(kind=8)       :: cn
      logical            :: match, cl
      character(len=100), dimension(100) :: temp

     ! default settings for logicals
     settings%avg_pixels = .false.
     settings%usegridfile = .false.
     settings%landonly = .false.
     settings%opmode_sel = -1

     ! open file
      open (100, file = trim (filename), status = 'old', iostat=ierr)
      if(ierr.gt.0) then
        write (*,*) 'ERROR: cannot find SETTINGS'
        stop
      endif

      ! loop over lines
      read_loop: do
        read (100, fmt='(A)', iostat=ierr) line

        ! first check if line exists or eof was reached
        if ( ierr.gt.0 ) then
          write (*,*) 'ERROR: check ', filename
          exit read_loop
        else if ( ierr.lt.0 ) then
          exit read_loop
        else

          ! omit comments ("#") and empty lines
          if ( index (trim (line), "#") .eq. 1 ) cycle read_loop
          if ( len   (trim (line))      .eq. 0 ) cycle read_loop

          ! Path and file settings
          identifier = "satellite:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%satellite = cc
          identifier = "satname:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%satname = cc
          identifier = "path_flexpart:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%path_flexpart = cc
          identifier = "path_options:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%path_options = cc
          identifier = "path_output:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%path_output = cc
          identifier = "path_ohfield:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%path_ohfield = cc
          identifier = "path_obs:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%path_obs = cc
          identifier = "path_obsout:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%path_obsout = cc
          identifier = "suffix:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%suffix = cc
          identifier = "file_avail:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%file_avail = cc
          identifier = "file_availnest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%file_availnest = cc
          identifier = "windfield:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%windfield = cc

          ! satellite variables
          identifier = "landonly:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%landonly = cl
          identifier = "opmode_sel:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%opmode_sel = cn
          identifier = "nmeas_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%nmeas_name = cc
          identifier = "npixel_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%npixel_name = cc
          identifier = "ncoord_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ncoord_name = cc
          identifier = "nlevel_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%nlevel_name = cc
          identifier = "oper_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%oper_name = cc
          identifier = "time_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%time_name = cc
          identifier = "tdel_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%tdel_name = cc
          identifier = "qa_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%qa_name = cc
          identifier = "qa_cutoff:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%qa_cutoff = cn
          identifier = "vmr_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%vmr_name = cc
          identifier = "verr_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%verr_name = cc
          identifier = "lat_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lat_name = cc
          identifier = "lon_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lon_name = cc
          identifier = "ak_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ak_name = cc
          identifier = "apri_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%apri_name = cc
          identifier = "dryair_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dryair_name = cc
          identifier = "psurf_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%psurf_name = cc
          identifier = "pdel_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%pdel_name = cc
          identifier = "surfclass_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%surfclass_name = cc
          identifier = "landfrac_name:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%landfrac_name = cc

          ! COMMAND settings
          identifier = "datei:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%datei = int(cn)
          identifier = "datef:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%datef = int(cn)
          identifier = "outrate:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outrate = int(cn)
          identifier = "outaverage:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outaverage = int(cn)
          identifier = "outsample:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outsample = int(cn)
          identifier = "ind_source:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ind_source = int(cn)
          identifier = "ind_receptor:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ind_receptor = int(cn)
          identifier = "lnested:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%lnested = int(cn)
          identifier = "linit_cond:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%linit_cond = int(cn)

          ! OUTGRID settings 
          identifier = "outlonleft:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outlonleft = real(cn)
          identifier = "outlatlower:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outlatlower = real(cn)
          identifier = "numxgrid:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%numxgrid = real(cn)
          identifier = "numygrid:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%numygrid = real(cn)
          identifier = "dxout:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dxout = real(cn)
          identifier = "dyout:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dyout = real(cn)
          identifier = "outlonnest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outlonnest = real(cn)
          identifier = "outlatnest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%outlatnest = real(cn)
          identifier = "numxnest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%numxnest = real(cn)
          identifier = "numynest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%numynest = real(cn)
          identifier = "dxoutnest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dxoutnest = real(cn)
          identifier = "dyoutnest:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dyoutnest = real(cn)
          identifier = "zlevel:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) then
            call parse_string (cc, ",", temp(:), n)
            if( .not.allocated(settings%zlevel) ) allocate( settings%zlevel(n) )
            settings%numzgrid = n
            do i = 1, n
              read(temp(i),*) settings%zlevel(i)
            enddo
          endif

          ! RELEASES settings
          identifier = "species:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%species = cc
          identifier = "ageclass:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%ageclass = int(cn)
          identifier = "npart:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%npart = int(cn)
          identifier = "mass:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%mass = int(cn)

          ! AVERAGING settings
          identifier = "avg_pixels:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%avg_pixels = cl
          identifier = "usegridfile:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%usegridfile = cl
          identifier = "dmin:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%dmin = cn
          identifier = "nsteps:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%nsteps = cn
          identifier = "cutoff:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%cutoff = cn
          identifier = "nmin:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%nmin = int(cn)
          identifier = "filegrid:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) settings%filegrid = cc

        endif

      end do read_loop

    end subroutine read_settings

    ! --------------------------------------------------

end module mod_settings

